﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;
using System.Diagnostics;

namespace LinqAsParallelOrdering {
	class Program {
		// A lock to protect "Console.Write" critical sections. Console.Write is thread-safe itself locking its own internal lock every time,
		// however whole Console class is not stateless (we are here modifing the ForegroundColor substate), so it cannot be safely used in a multithreaded environment.
		// In a real scenario (without Console writes) the lock is not necessary for correct functionality!
		static object consoleLock = new object();

		static int parallelFs = 0;
		static int parallelXs = 0;
		static int inF = 0;
		static int inX = 0;

		static IEnumerable<Random> GetRepeatableRandoms(int seed) {
			while (true) {
				yield return new Random(seed);
			}
		}

		static bool f(Tuple<int, int> value) {
			if (Interlocked.Increment(ref inF) > 1) {
				Interlocked.Increment(ref parallelFs);
			}
			lock (consoleLock) Console.Write("{0}f({1}) ", Thread.CurrentThread.ManagedThreadId, value.Item1);
			Thread.Sleep(value.Item2);
			lock (consoleLock) {
				Console.ForegroundColor = ConsoleColor.White;
				Console.Write("{0}r({1}) ", Thread.CurrentThread.ManagedThreadId, value.Item1);
				Console.ResetColor();
			}
			Interlocked.Decrement(ref inF);
			return true;
		}

		static bool X(Tuple<int, int> value) {
			if (Interlocked.Increment(ref inX) > 1) {
				Interlocked.Increment(ref parallelXs);
			}
			lock (consoleLock) Console.Write("{0}X({1}) ", Thread.CurrentThread.ManagedThreadId, value.Item1);
			Thread.Sleep(5);
			Interlocked.Decrement(ref inX);

			return true;
		}

		static void Process(int item) {
			lock (consoleLock) {
				Console.ForegroundColor = ConsoleColor.Green;
				Console.Write("{0}p({1}) ", Thread.CurrentThread.ManagedThreadId, item);
				Console.ResetColor();
			}
			Thread.Sleep(50);
		}

		static void WriteInRed(string text) {
			lock (consoleLock) {
				Console.ForegroundColor = ConsoleColor.Red;
				Console.Write(text);
				Console.ResetColor();
			}
		}

		static void Main(string[] args) {

			var source = from r in GetRepeatableRandoms(42).Take(1) from x in Enumerable.Range(1, 100) select new Tuple<int, int>(x, r.Next(200));
			source = from x in source where X(x) select x;

			{
				Console.WriteLine("Standard LINQ to Objects single-threaded query:");
				Console.WriteLine("foreach from x in source where f(x) select x.Item1");
				var pq = from x in source where f(x) select x.Item1;

				WriteInRed(string.Format("{0}Main-sleeping ", Thread.CurrentThread.ManagedThreadId));
				Thread.Sleep(500);
				WriteInRed(string.Format("{0}Main-enumStart ", Thread.CurrentThread.ManagedThreadId));

				var sw = Stopwatch.StartNew();
				foreach (var i in pq) {
					Process(i);
				}
				sw.Stop();

				WriteInRed(string.Format("{0}Main-enumEnd(time {1}) ", Thread.CurrentThread.ManagedThreadId, sw.Elapsed));
			}
			Console.WriteLine();
			Console.WriteLine($"{parallelFs}-times in f() while in other f() + {parallelXs}-times in X() while in other X()"); parallelFs = 0; parallelXs = 0;
			Console.WriteLine(); Console.WriteLine();
			Console.Write("Press ENTER to show PLINQ ...");
			Console.ReadLine();

			{
				Console.WriteLine("foreach from x in source.AsParallel() where f(x) select x.Item1");
				var pq = from x in source.AsParallel() where f(x) select x.Item1;

				WriteInRed(string.Format("{0}Main-sleeping ", Thread.CurrentThread.ManagedThreadId));
				Thread.Sleep(500);
				WriteInRed(string.Format("{0}Main-enumStart ", Thread.CurrentThread.ManagedThreadId));

				var sw = Stopwatch.StartNew();
				foreach (var i in pq) {
					Process(i);
				}
				sw.Stop();

				WriteInRed(string.Format("{0}Main-enumEnd(time {1}) ", Thread.CurrentThread.ManagedThreadId, sw.Elapsed));
			}
			Console.WriteLine();
			Console.WriteLine($"{parallelFs}-times in f() while in other f() + {parallelXs}-times in X() while in other X()"); parallelFs = 0; parallelXs = 0;
			Console.WriteLine(); Console.WriteLine();
			Console.Write("Press ENTER to show more complex examples ...");
			Console.ReadLine();

			{
				Console.WriteLine("foreach from x in source.AsParallel().AsOrdered() where f(x) select x.Item1");
				var pq = from x in source.AsParallel().AsOrdered() where f(x) select x.Item1;

				WriteInRed(string.Format("{0}Main-sleeping ", Thread.CurrentThread.ManagedThreadId));
				Thread.Sleep(500);
				WriteInRed(string.Format("{0}Main-enumStart ", Thread.CurrentThread.ManagedThreadId));

				var sw = Stopwatch.StartNew();
				foreach (var i in pq) {
					Process(i);
				}
				sw.Stop();

				WriteInRed(string.Format("{0}Main-enumEnd(time {1}) ", Thread.CurrentThread.ManagedThreadId, sw.Elapsed));
			}
			Console.WriteLine();
			Console.WriteLine($"{parallelFs}-times in f() while in other f() + {parallelXs}-times in X() while in other X()"); parallelFs = 0; parallelXs = 0;
			Console.WriteLine(); Console.WriteLine();
			Console.Write("Press ENTER to show MergeOptions ...");
			Console.ReadLine();

			{
				Console.WriteLine("foreach from x in source.AsParallel().WithMergeOptions(ParallelMergeOptions.NotBuffered).AsOrdered() where f(x) select x.Item1");
				var pq = from x in source.AsParallel().AsOrdered().WithMergeOptions(ParallelMergeOptions.NotBuffered) where f(x) select x.Item1;

				WriteInRed(string.Format("{0}Main-sleeping ", Thread.CurrentThread.ManagedThreadId));
				Thread.Sleep(500);
				WriteInRed(string.Format("{0}Main-enumStart ", Thread.CurrentThread.ManagedThreadId));

				var sw = Stopwatch.StartNew();
				foreach (var i in pq) {
					Process(i);
				}
				sw.Stop();

				WriteInRed(string.Format("{0}Main-enumEnd(time {1}) ", Thread.CurrentThread.ManagedThreadId, sw.Elapsed));
			}
			Console.WriteLine();
			Console.WriteLine($"{parallelFs}-times in f() while in other f() + {parallelXs}-times in X() while in other X()"); parallelFs = 0; parallelXs = 0;
			Console.WriteLine(); Console.WriteLine();

			{
				Console.WriteLine("foreach from x in source.AsParallel().WithMergeOptions(ParallelMergeOptions.NotBuffered).AsUnordered() where f(x) select x.Item1");
				var pq = from x in source.AsParallel().AsUnordered().WithMergeOptions(ParallelMergeOptions.NotBuffered) where f(x) select x.Item1;

				WriteInRed(string.Format("{0}Main-sleeping ", Thread.CurrentThread.ManagedThreadId));
				Thread.Sleep(500);
				WriteInRed(string.Format("{0}Main-enumStart ", Thread.CurrentThread.ManagedThreadId));

				var sw = Stopwatch.StartNew();
				foreach (var i in pq) {
					Process(i);
				}
				sw.Stop();

				WriteInRed(string.Format("{0}Main-enumEnd(time {1}) ", Thread.CurrentThread.ManagedThreadId, sw.Elapsed));
			}
			Console.WriteLine();
			Console.WriteLine($"{parallelFs}-times in f() while in other f() + {parallelXs}-times in X() while in other X()"); parallelFs = 0; parallelXs = 0;
			Console.WriteLine(); Console.WriteLine();
			Console.Write("Press ENTER to show ForAll vs. ForEach ...");
			Console.ReadLine();

			{
				Console.WriteLine("(from x in source.AsParallel().AsOrdered() where f(x) select x.Item1).ForAll");
				var pq = from x in source.AsParallel().AsOrdered() where f(x) select x.Item1;

				WriteInRed(string.Format("{0}Main-sleeping ", Thread.CurrentThread.ManagedThreadId));
				Thread.Sleep(500);
				WriteInRed(string.Format("{0}Main-enumStart ", Thread.CurrentThread.ManagedThreadId));

				var sw = Stopwatch.StartNew();
				pq.ForAll(Process);
				sw.Stop();

				WriteInRed(string.Format("{0}Main-enumEnd(time {1}) ", Thread.CurrentThread.ManagedThreadId, sw.Elapsed));
			}
			Console.WriteLine();
			Console.WriteLine($"{parallelFs}-times in f() while in other f() + {parallelXs}-times in X() while in other X()"); parallelFs = 0; parallelXs = 0;
			Console.WriteLine(); Console.WriteLine();

			{
				Console.WriteLine("(from x in source.AsParallel().AsOrdered() where f(x) select x.Item1).ForAll");
				var pq = from x in source.AsParallel().AsUnordered() where f(x) select x.Item1;

				WriteInRed(string.Format("{0}Main-sleeping ", Thread.CurrentThread.ManagedThreadId));
				Thread.Sleep(500);
				WriteInRed(string.Format("{0}Main-enumStart ", Thread.CurrentThread.ManagedThreadId));

				var sw = Stopwatch.StartNew();
				pq.ForAll(Process);
				sw.Stop();

				WriteInRed(string.Format("{0}Main-enumEnd(time {1}) ", Thread.CurrentThread.ManagedThreadId, sw.Elapsed));
			}
			Console.WriteLine();
			Console.WriteLine($"{parallelFs}-times in f() while in other f() + {parallelXs}-times in X() while in other X()"); parallelFs = 0; parallelXs = 0;
			Console.WriteLine(); Console.WriteLine();
			
		}
	}
}
